@@ -30,46 +30,48 @@ use 5.10.0;
 use strict;
 use warnings;
 
-use URI;
 use IO::Select;
 use IO::Pipe;
 use IO::Socket::UNIX;
 use POSIX;
 use Time::HiRes qw(setitimer ITIMER_REAL);
+use URI;
 
-use fields qw(_config _progress _worker_pid _server_socket _parent_pipe _server_socket_path);
+use Cupt::LValueFields qw(_config _progress _worker_pid _server_socket
+		_parent_pipe _server_socket_path);
 
 use Cupt::Core;
-use Cupt::Download::Methods::Curl;
-use Cupt::Download::Methods::File;
+use Cupt::Download::Method;
 
 sub __my_write_socket ($@) {
 	my $socket = shift;
-	defined $socket or myinternaldie("bad socket parameter");
+	defined $socket or myinternaldie('bad socket parameter');
 
 	my $string = join(chr(1), @_);
 	my $len = length($string);
-	my $packed_len = pack("S", $len);
+	my $packed_len = pack('S', $len);
 
 	syswrite($socket, ($packed_len . $string)) or
-			myinternaldie("write to socket failed: $!")
+			mydie('write to socket failed: %s', $!);
+	return;
 }
 
 sub __my_read_socket ($) {
 	my $socket = shift;
-	defined $socket or myinternaldie("bad socket parameter");
+	defined $socket or myinternaldie('bad socket parameter');
 
 	my $string;
 
 	my $read_result = sysread($socket, my $packed_len, 2);
-	defined $read_result or myinternaldie("read from socket failed: $!");
+	defined $read_result or mydie('read from socket failed: %s', $!);
 
 	if ($read_result == 0) {
 		$string = 'eof';
 	} else {
-		my ($len) = unpack("S", $packed_len);
+		my ($len) = unpack('S', $packed_len);
 		# don't use anything but sysread here, as we use select() on sockets
-		sysread $socket, $string, $len;
+		sysread($socket, $string, $len) or
+				mydie('read from socket failed: %s', ($! // 'end of file'));
 	}
 
 	return split(chr(1), $string, -1);
@@ -92,44 +94,49 @@ I<progress> - reference to object of subclass of L<Cupt::Download::Progress|Cupt
 sub new ($$$) {
 	my $class = shift;
 	my $self;
-	$self = fields::new($class);
-	$self->{_config} = shift;
-	$self->{_progress} = shift;
+	$self = bless [] => $class;
+	$self->_config = shift;
+	$self->_progress = shift;
 
-	# main socket
-	$self->{_server_socket_path} = "/tmp/cupt-downloader-$$";
-	unlink($self->{_server_socket_path}); # intentionally ignore errors
-	$self->{_server_socket} = new IO::Socket::UNIX(
-			Local => $self->{_server_socket_path}, Listen => SOMAXCONN, Type => SOCK_STREAM);
-	defined $self->{_server_socket} or
-			mydie("unable to open server socket on file '%s': %s", $self->{_server_socket_path}, $!);
-	
-	$self->{_parent_pipe} = new IO::Pipe() //
-			mydie("unable to open parent pipe: %s", $!);
-
-	my $pid = fork();
-	defined $pid or
-			mydie("unable to create download worker process: %s", $!);
-
-	if ($pid) {
-		# this is a main process
-		$self->{_worker_pid} = $pid;
-		$self->{_parent_pipe}->writer();
-		$self->{_parent_pipe}->autoflush(1);
+	if ($self->_config->get_bool('cupt::worker::simulate')) {
 		return $self;
 	} else {
-		# this is background worker process
-		$self->{_parent_pipe}->reader();
-		$self->_worker();
+		# main socket
+		$self->_server_socket_path = "/tmp/cupt-downloader-$$";
+		unlink($self->_server_socket_path); ## no critic (RequireCheckedSyscalls)
+		$self->_server_socket = IO::Socket::UNIX->new(
+				Local => $self->_server_socket_path, Listen => SOMAXCONN, Type => SOCK_STREAM);
+		defined $self->_server_socket or
+				mydie("unable to open server socket on file '%s': %s", $self->_server_socket_path, $!);
+
+		$self->_parent_pipe = IO::Pipe->new() //
+				mydie('unable to open parent pipe: %s', $!);
+
+		my $pid = fork();
+		defined $pid or
+				mydie('unable to create download worker process: %s', $!);
+
+		if ($pid) {
+			# this is a main process
+			$self->_worker_pid = $pid;
+			$self->_parent_pipe->writer();
+			$self->_parent_pipe->autoflush(1);
+			return $self;
+		} else {
+			# this is background worker process
+			$self->_parent_pipe->reader();
+			$self->_worker();
+		}
 	}
+	return;
 }
 
 sub _worker ($) {
 	my ($self) = @_;
 
-	my $debug = $self->{_config}->var('debug::downloader');
+	my $debug = $self->_config->get_bool('debug::downloader');
 
-	mydebug("download worker process started") if $debug;
+	mydebug('download worker process started') if $debug;
 
 	# { $uri => $result }
 	my %done_downloads;
@@ -145,7 +152,7 @@ sub _worker ($) {
 	# { $uri => $filename }
 	my %target_filenames;
 
-	my $max_simultaneous_downloads_allowed = $self->{_config}->var('cupt::downloader::max-simultaneous-downloads');
+	my $max_simultaneous_downloads_allowed = $self->_config->get_number('cupt::downloader::max-simultaneous-downloads');
 	pipe(my $worker_reader, my $worker_writer) or
 			mydie("unable to open worker's own pair of sockets: %s", $!);
 	$worker_writer->autoflush(1);
@@ -153,12 +160,15 @@ sub _worker ($) {
 	my $exit_flag = 0;
 
 	# setting progress ping timer
-	$SIG{ALRM} = sub { __my_write_socket($worker_writer, 'progress', '', 'ping') };
+	local $SIG{ALRM} = sub { __my_write_socket($worker_writer, 'progress', '', 'ping') };
 	setitimer(ITIMER_REAL, 0.25, 0.25);
 
-	my @persistent_sockets = ($worker_reader, $self->{_parent_pipe}, $self->{_server_socket});
+	my @persistent_sockets = ($worker_reader, $self->_parent_pipe, $self->_server_socket);
 	my @runtime_sockets;
-	while (!$exit_flag) {
+
+	# while caller may set exit flag, we should continue processing as long as
+	# something is pending in internal queue
+	while (!$exit_flag || IO::Select->new($worker_reader)->can_read(0)) {
 		# periodic cleaning of sockets which were accept()'ed
 		@runtime_sockets = grep { $_->opened } @runtime_sockets;
 
@@ -167,12 +177,12 @@ sub _worker ($) {
 
 		foreach my $socket (@ready) {
 			next if not $socket->opened;
-			if ($socket eq $self->{_server_socket}) {
+			if ($socket eq $self->_server_socket) {
 				# a new connection appeared
 				$socket = $socket->accept();
 				defined $socket or
-						mydie("unable to accept new socket connection: %s", $!);
-				mydebug("accepted new connection") if $debug;
+						mydie('unable to accept new socket connection: %s', $!);
+				mydebug('accepted new connection') if $debug;
 				push @runtime_sockets, $socket;
 			}
 			my @params = __my_read_socket($socket);
@@ -184,13 +194,13 @@ sub _worker ($) {
 			my $proceed_next_download = 0;
 			given ($command) {
 				when ('exit') {
-					mydebug("exit scheduled") if $debug;
+					mydebug('exit scheduled') if $debug;
 					$exit_flag = 1;
 				}
 				when ('eof') {
 					# the current socket reported EOF
-					mydebug("eof has been reported") if $debug;
-					close($socket) or mydie("unable to close socket: %s", $!);
+					mydebug('eof has been reported') if $debug;
+					close($socket) or mydie('unable to close socket: %s', $!);
 				}
 				when ('download') {
 					# new query appeared
@@ -215,9 +225,9 @@ sub _worker ($) {
 
 					# clean after child
 					close($active_downloads{$uri}->{performer_reader}) or
-							mydie("unable to close performer socket: %s", $!);
+							mydie('unable to close performer socket: %s', $!);
 
-					waitpid($active_downloads{$uri}->{pid}, 0);
+					waitpid($active_downloads{$uri}->{pid}, 0) ## no critic (RequireCheckedSyscalls);
 				}
 				when ('done-ack') {
 					# this is final ACK from download with final result
@@ -232,24 +242,25 @@ sub _worker ($) {
 					delete $active_downloads{$uri};
 					$done_downloads{$uri} = $result;
 
-					mydebug("started checking pending queue") if $debug;
+					mydebug('started checking pending queue') if $debug;
 					do { # answering on duplicated requests if any
 						my $is_duplicated_download = 1;
 						my @new_pending_downloads;
 
 						foreach my $ref_pending_download (@pending_downloads) {
-							(my $uri, $waiter_socket) = @$ref_pending_download;
-							if (exists $done_downloads{$uri}) {
-								mydebug("final download result for duplicated request: '$uri': $result") if $debug;
-								__my_write_socket($waiter_socket, $uri, $result, $is_duplicated_download);
-								close($waiter_socket);
+							(my $pending_uri, $waiter_socket) = @$ref_pending_download;
+							if (exists $done_downloads{$pending_uri}) {
+								mydebug("final download result for duplicated request: '$pending_uri': $result") if $debug;
+								__my_write_socket($waiter_socket, $pending_uri, $result, $is_duplicated_download);
+								close($waiter_socket) or
+										mydie("unable to close waiter socket for uri '%s'", $pending_uri);
 							} else {
 								push @new_pending_downloads, $ref_pending_download;
 							}
 						}
 						@pending_downloads = @new_pending_downloads;
 					};
-					mydebug("finished checking pending queue") if $debug;
+					mydebug('finished checking pending queue') if $debug;
 
 					# update progress
 					__my_write_socket($worker_writer, 'progress', $uri, 'done', $result);
@@ -268,16 +279,18 @@ sub _worker ($) {
 							# so, this download don't make sense
 							$filename = $target_filenames{$uri};
 							# rest in peace, young process
-							kill SIGTERM, $active_downloads{$uri}->{pid};
+							kill(SIGTERM, $active_downloads{$uri}->{'pid'}) or
+									mywarn('unable to kill process %u: %s', $active_downloads{$uri}->{'pid'}, $!);
 							# process it as failed
-							my $error_string = sprintf __("invalid size: expected '%u', got '%u'"),
+							my $error_string = sprintf __("invalid size: expected '%u', got '%d'"),
 									$download_sizes{$uri}, $expected_size;
 							__my_write_socket($worker_writer, 'done', $uri, $error_string);
-							unlink $filename;
+							unlink $filename or
+									mywarn("unable to delete file '%s': %s", $filename, $!);
 						}
 					} else {
 						# update progress
-						$self->{_progress}->progress($uri, $action, @params);
+						$self->_progress->progress($uri, $action, @params);
 					}
 				}
 				when ('pop-download') {
@@ -289,12 +302,12 @@ sub _worker ($) {
 					}
 				}
 				when ('set-long-alias') {
-					$self->{_progress}->set_long_alias_for_uri(@params);
+					$self->_progress->set_long_alias_for_uri(@params);
 				}
 				when ('set-short-alias') {
-					$self->{_progress}->set_short_alias_for_uri(@params);
+					$self->_progress->set_short_alias_for_uri(@params);
 				}
-				default { myinternaldie("download manager: invalid worker command"); }
+				default { myinternaldie('download manager: invalid worker command'); }
 			}
 
 			$proceed_next_download or next;
@@ -329,11 +342,11 @@ sub _worker ($) {
 			# filling the active downloads hash
 			$active_downloads{$uri}->{waiter_socket} = $waiter_socket;
 
-			my $performer_pipe = new IO::Pipe() //
+			my $performer_pipe = IO::Pipe->new() //
 					mydie("unable to open performer's pair of sockets: %s", $!);
 
 			my $download_pid = fork();
-			$download_pid // mydie("unable to fork: %s", $!);
+			$download_pid // mydie('unable to fork: %s', $!);
 			$active_downloads{$uri}->{pid} = $download_pid;
 
 			if ($download_pid) {
@@ -342,7 +355,7 @@ sub _worker ($) {
 				$active_downloads{$uri}->{performer_reader} = $performer_pipe;
 			} else {
 				# background downloader process
-				$SIG{TERM} = sub { POSIX::_exit(0) };
+				local $SIG{TERM} = sub { POSIX::_exit(0) };
 
 				$performer_pipe->writer();
 				$performer_pipe->autoflush(1);
@@ -355,7 +368,7 @@ sub _worker ($) {
 				__my_write_socket($performer_pipe, @progress_message);
 
 				my $result = $self->_download($uri, $filename, $performer_pipe);
-				myinternaldie("a download method returned undefined result") if not defined $result;
+				myinternaldie('a download method returned undefined result') if not defined $result;
 
 				__my_write_socket($performer_pipe, 'done', $uri, $result);
 
@@ -364,32 +377,43 @@ sub _worker ($) {
 		}
 	}
 	# disabling timer
-	$SIG{ALRM} = sub {};
+	local $SIG{ALRM} = sub {};
 	setitimer(ITIMER_REAL, 0, 0);
 	# finishing progress
-	$self->{_progress}->finish();
+	$self->_progress->finish();
 
 	close($worker_reader) or mydie("unable to close worker's own reader socket: %s", $!);
 	close($worker_writer) or mydie("unable to close worker's own writer socket: %s", $!);
-	mydebug("download worker process finished") if $debug;
+	mydebug('download worker process finished') if $debug;
 	POSIX::_exit(0);
+	return;
 }
 
 sub DESTROY {
 	my ($self) = @_;
-	# shutdowning worker thread
-	__my_write_socket($self->{_parent_pipe}, 'exit');
-	waitpid($self->{_worker_pid}, 0);
 
-	# cleaning parent sockets
-	close($self->{_parent_reader}) or mydie("unable to close parent reader socket: %s", $!);
-	close($self->{_parent_writer}) or mydie("unable to close parent writer socket: %s", $!);
+	unless ($self->_config->get_bool('cupt::worker::simulate')) {
+		# shutdowning worker thread
+		__my_write_socket($self->_parent_pipe, 'exit');
+		waitpid($self->_worker_pid, 0); ## no critic (RequireCheckedSyscalls)
 
-	# cleaning server socket
-	close($self->{_server_socket}) or
-			mydie("unable to close server socket on file '%s': %s", $self->{_server_socket_path}, $!);
-	unlink($self->{_server_socket_path}) or
-			mydie("unable to delete server socket file '%s': %s", $self->{_server_socket_path}, $!);
+		# cleaning parent sockets
+		close($self->_parent_pipe) or mydie('unable to close parent writer socket: %s', $!);
+
+		# cleaning server socket
+		close($self->_server_socket) or
+				mydie("unable to close server socket on file '%s': %s", $self->_server_socket_path, $!);
+		unlink($self->_server_socket_path) or
+				mydie("unable to delete server socket file '%s': %s", $self->_server_socket_path, $!);
+	}
+	return;
+}
+
+sub _get_uri_priority {
+	my ($self, $uri) = @_;
+
+	my $protocol = URI->new($uri)->scheme();
+	return $self->_config->get_number("cupt::downloader::protocols::${protocol}::priority") // 100;
 }
 
 =head2 download
@@ -403,14 +427,34 @@ Parameters:
 
 Sequence of hash entries with the following fields:
 
-I<uris> - array of mirror URIs to download, mandatory
+=over
 
-I<filename> - target filename, mandatory
+=item uri-entries
 
-I<post-action> - reference to subroutine that will be called in case of
+array of I<URI entry>s, mandatory
+
+I<URI entry> - hash entry with keys:
+
+uri - mirror URIs to download, mandatory
+
+short-alias - short alias for this URI, optional
+
+long-alias - long alias for this URI, optional
+
+=item filename
+
+target filename, mandatory
+
+=item post-action
+
+reference to subroutine that will be called in case of
 successful download, optional
 
-I<size> - fixed size for target, will be used in sanity checks, optional
+=item size
+
+fixed size for target, will be used in sanity checks, optional
+
+=back
 
 Returns:
 
@@ -420,12 +464,20 @@ Example:
 
   my $download_manager = new Cupt::Download::Manager;
   $download_manager->download(
-    { 'uris' => [ 'http://www.en.debian.org' ], 'filename' => '/tmp/en.html' },
-    { 'uris' => [ 'http://www.ru.debian.org' ], 'filename' => '/tmp/ru.html', 'post-action' => \&checker },
-    { 'uris' => [ 'http://www.ua.debian.org' ], 'filename' => '/tmp/ua.html', 'size' => 10254 }
-    { 'uris' => [
-        'http://ftp.de.debian.org/debian/pool/main/n/nlkt/nlkt_0.3.2.1-2_amd64.deb',
-        'http://ftp.es.debian.org/debian/pool/main/n/nlkt/nlkt_0.3.2.1-2_amd64.deb'
+    { 'uri-entries' => [ { 'uri' => 'http://www.en.debian.org' } ], 'filename' => '/tmp/en.html' },
+    { 'uri-entries' => [ { 'uri' => 'http://www.ru.debian.org' } ], 'filename' => '/tmp/ru.html', 'post-action' => \&checker },
+    { 'uri-entries' => [ { 'uri' => 'http://www.ua.debian.org' } ], 'filename' => '/tmp/ua.html', 'size' => 10254 }
+    { 'uri-entries' => [
+        {
+          'uri' => 'http://ftp.de.debian.org/debian/pool/main/n/nlkt/nlkt_0.3.2.1-2_amd64.deb',
+          'long-alias' => 'ftp.de.debian.org/debian sid/main nlkt 0.3.2.1-2',
+          'short-alias' => 'nlkt',
+        },
+        {
+          'uri' => 'http://ftp.ua.debian.org/debian/pool/main/n/nlkt/nlkt_0.3.2.1-2_amd64.deb',
+          'long-alias' => 'ftp.ua.debian.org/debian sid/main nlkt 0.3.2.1-2',
+          'short-alias' => 'nlkt',
+        },
       ], 'filename' => '/var/cache/apt/archives/nlkt_0.3.2.1-2_amd64.deb' }
   );
 
@@ -434,26 +486,42 @@ Example:
 sub download ($@) {
 	my $self = shift;
 
-	# { $filename => { 'uris' => [ $uri... ], 'size' => $size, 'checker' => $checker }... }
+	if ($self->_config->get_bool('cupt::worker::simulate')) {
+		foreach my $ref_download_entry (@_) {
+			my @uris = map { $_->{'uri'} } @{$ref_download_entry->{'uri-entries'}};
+			mysimulate('downloading: %s', join(' | ', @uris));
+		}
+		return 0;
+	}
+
 	my %download_entries;
 
 	# { $uri => $filename }
 	my %waiters;
 
-	my $socket = new IO::Socket::UNIX($self->{_server_socket_path});
-	defined $socket or mydie("unable to open download socket: %s", $!);
+	my $socket = IO::Socket::UNIX->new($self->_server_socket_path);
+	defined $socket or mydie('unable to open download socket: %s', $!);
 
 	my $sub_schedule_download = sub {
 		my ($filename) = @_;
 
-		$download_entries{$filename}->{'current_uri'} = shift @{$download_entries{$filename}->{'uris'}};
+		my $current_uri_entry = shift @{$download_entries{$filename}->{'uris'}};
+		$download_entries{$filename}->{'current_uri'} = $current_uri_entry->{'uri'};
 		my $uri = $download_entries{$filename}->{'current_uri'};
 		my $size = $download_entries{$filename}->{'size'};
 		my $checker = $download_entries{$filename}->{'checker'};
+		my $short_alias = $current_uri_entry->{'short-alias'};
+		my $long_alias = $current_uri_entry->{'long-alias'};
 
 		if (defined $size) {
 			__my_write_socket($socket, 'set-download-size', $uri, $size);
 		}
+		if (defined $short_alias) {
+			__my_write_socket($socket, 'set-short-alias', $uri, $short_alias);
+		}
+		if (defined $long_alias) {
+			__my_write_socket($socket, 'set-long-alias', $uri, $long_alias);
+		}
 		__my_write_socket($socket, 'download', $uri, $filename);
 
 		$waiters{$uri} = $filename;
@@ -463,10 +531,12 @@ sub download ($@) {
 	while (scalar @_) {
 		# extract next entry
 		my $ref_entry = shift;
-		my $ref_uris= $ref_entry->{'uris'};
 		my $filename = $ref_entry->{'filename'};
 
-		$download_entries{$filename}->{'uris'} = $ref_uris;
+		# sorting by protocols' priorities
+		$download_entries{$filename}->{'uris'} = [ sort {
+					$self->_get_uri_priority($b->{'uri'}) <=> $self->_get_uri_priority($a->{'uri'})
+				} @{$ref_entry->{'uri-entries'}} ];
 		# may be undef
 		$download_entries{$filename}->{'size'} = $ref_entry->{'size'};
 		# may be undef
@@ -479,16 +549,21 @@ sub download ($@) {
 	my $result = 0;
 	while (scalar keys %waiters) {
 		my ($uri, $error_string, $is_duplicated_download) = __my_read_socket($socket);
+		defined $uri or myinternaldie('received undefined uri');
 		my $filename = $waiters{$uri};
+		defined $filename or myinternaldie("received wrong uri $uri");
 		my $sub_post_action = $download_entries{$filename}->{'checker'};
 
 		# delete entry from list
 		delete $waiters{$uri};
 
-		if (!$error_string && defined $sub_post_action && not $is_duplicated_download) {
+		if (not $error_string and defined $sub_post_action and not $is_duplicated_download) {
 			# download seems to be done well, but we also have external checker specified
 			# but do this only if this file wasn't post-processed before
 			$error_string = $sub_post_action->();
+			if (not defined $error_string) {
+				myinternaldie("received undefined post-action result for $uri");
+			}
 		}
 
 		if (not $is_duplicated_download) {
@@ -509,59 +584,20 @@ sub download ($@) {
 		}
 	}
 
-	close($socket) or mydie("unable to close download socket: %s", $!);
+	close($socket) or mydie('unable to close download socket: %s', $!);
 
 	# finish
 	return $result;
 }
 
-=head2 set_short_alias_for_uri
-
-method, forwards params to underlying download progress
-
-=cut
-
-sub set_short_alias_for_uri {
-	my ($self, @params) = @_;
-	__my_write_socket($self->{_parent_pipe}, 'set-short-alias', @params);
-}
-
-=head2 set_long_alias_for_uri
-
-method, forwards params to underlying download progress
-
-=cut
-
-sub set_long_alias_for_uri {
-	my ($self, @params) = @_;
-	__my_write_socket($self->{_parent_pipe}, 'set-long-alias', @params);
-}
-
 sub _download ($$$) {
 	my ($self, $uri, $filename, $socket) = @_;
 
-	my %protocol_handlers = (
-		'http' => 'Curl',
-		'ftp' => 'Curl',
-		'https' => 'Curl',
-		'file' => 'File',
-	);
-	my $protocol = URI->new($uri)->scheme();
-	my $handler_name = $protocol_handlers{$protocol} // 
-			return sprintf __("no protocol download handler defined for %s"), $protocol;
-
-	my $handler;
-	{
-		no strict 'subs';
-		# create handler by name
-		$handler = "Cupt::Download::Methods::$handler_name"->new();
-	}
-
-	# download the file
 	my $sub_callback = sub {
 		__my_write_socket($socket, 'progress', $uri, @_);
 	};
-	my $result = $handler->perform($self->{_config}, $uri, $filename, $sub_callback);
+	my $download_method = Cupt::Download::Method->new();
+	my $result = $download_method->perform($self->_config, $uri, $filename, $sub_callback);
 	return $result;
 }
 
